(* generated by dune *)

(** @canonical Dune_re.Automata *)
module Automata = Dune_re__Automata

(** @canonical Dune_re.Category *)
module Category = Dune_re__Category

(** @canonical Dune_re.Color_map *)
module Color_map = Dune_re__Color_map

(** @canonical Dune_re.Core *)
module Core = Dune_re__Core

(** @canonical Dune_re.Cset *)
module Cset = Dune_re__Cset

(** @canonical Dune_re.Emacs *)
module Emacs = Dune_re__Emacs

(** @canonical Dune_re.Fmt *)
module Fmt = Dune_re__Fmt

(** @canonical Dune_re.Glob *)
module Glob = Dune_re__Glob

(** @canonical Dune_re.Group *)
module Group = Dune_re__Group

(** @canonical Dune_re.Pcre *)
module Pcre = Dune_re__Pcre

(** @canonical Dune_re.Perl *)
module Perl = Dune_re__Perl

(** @canonical Dune_re.Pmark *)
module Pmark = Dune_re__Pmark

(** @canonical Dune_re.Posix *)
module Posix = Dune_re__Posix

(** @canonical Dune_re.Re *)
module Re = Dune_re__Re

(** @canonical Dune_re.Str *)
module Str = Dune_re__Str

module Dune_re__ = struct end
[@@deprecated "this module is shadowed"]
